/* Mouse Position for X11 */
/* By Steven Lawrance */

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include "net_sourceforge_kjmouse_X11MousePosition.h"
#include <jawt.h>
#include <jawt_md.h>

/* Include our configuration */
#include "config.h"
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

/* Define our native pointer integer type (is there a standardized way to do this?) */
#if SIZEOF_CHAR_P == 4
  /* 32-bit CPU */
  #ifndef u_int32_t
  #define u_int32_t unsigned int
  #endif
  #define P_TYPE u_int32_t
#else
#if SIZEOF_CHAR_P == 8
  /* 64-bit CPU */
  #define P_TYPE u_int64_t
#else
  /* Unsupported CPU type */
  #error libkjmouse will only work on 32-bit and 64-bit CPUs at this time
#endif
#endif

#ifdef __cplusplus
extern "C" {
#endif

JNIEXPORT jboolean JNICALL Java_net_sourceforge_kjmouse_X11MousePosition_initialize (JNIEnv *env, jobject obj)
{
    /* Get the Java fields */
    Window root;
    Display *display;
    jfieldID x_id, y_id, display_id, root_id;
    jclass cls;
    (*env)->ExceptionClear(env);
    cls = (*env)->GetObjectClass(env, obj);
    x_id = (*env)->GetFieldID(env, cls, "x", "I");
    y_id = (*env)->GetFieldID(env, cls, "y", "I");
    display_id = (*env)->GetFieldID(env, cls, "display", "J");
    root_id = (*env)->GetFieldID(env, cls, "root", "I");

    /* Set x and y to zero */
    (*env)->SetIntField(env, obj, x_id, 0);
    (*env)->SetIntField(env, obj, y_id, 0);
    
    /* Open the Display and get the root Window */
    display = XOpenDisplay(0);
    if (!display) {
        (*env)->SetLongField(env, obj, display_id, 0);
        return 0;
    }
    root = XRootWindow(display, XDefaultScreen(display));
    
    /* Remember the display and root variables for later */
    (*env)->SetLongField(env, obj, display_id, ((P_TYPE) display));
    (*env)->SetIntField(env, obj, root_id, root);
    
    /* Return success */
    return 1;
}

JNIEXPORT void JNICALL Java_net_sourceforge_kjmouse_X11MousePosition_finalize (JNIEnv *env, jobject obj)
{
    /* Get the Java fields */
    jfieldID display_id;
    jclass cls;
    Display *display;
    (*env)->ExceptionClear(env);
    cls = (*env)->GetObjectClass(env, obj);
    display_id = (*env)->GetFieldID(env, cls, "display", "J");

    /* Load the value that we need */
    display = (Display *) ((P_TYPE) (*env)->GetLongField(env, obj, display_id));
    if (!display)
        return;

    /* Close the Display */    
    XCloseDisplay(display);
    
    /* Zero out display  */
    (*env)->SetLongField(env, obj, display_id, 0);
}

JNIEXPORT jboolean JNICALL Java_net_sourceforge_kjmouse_X11MousePosition_getMousePosition (JNIEnv *env, jobject obj)
{
    /* Get the Java fields */
    jfieldID x_id, y_id, display_id, root_id;
    jint prevX, prevY;
    jclass cls;
    Window root, RootIDRet, ChildIDRet;
    Display *display;
    unsigned int StateMask;
    int WinX, WinY;
    int XCoord, YCoord;
    (*env)->ExceptionClear(env);
    cls = (*env)->GetObjectClass(env, obj);
    x_id = (*env)->GetFieldID(env, cls, "x", "I");
    y_id = (*env)->GetFieldID(env, cls, "y", "I");
    display_id = (*env)->GetFieldID(env, cls, "display", "J");
    root_id = (*env)->GetFieldID(env, cls, "root", "I");

    /* Load the values that we need */
    prevX = (*env)->GetIntField(env, obj, x_id);
    prevY = (*env)->GetIntField(env, obj, y_id);
    display = (Display *) ((P_TYPE) (*env)->GetLongField(env, obj, display_id));
    root = (Window) (*env)->GetIntField(env, obj, root_id);
    if (!display)
        return 0;

    /* Get the current mouse cursor position */
    XQueryPointer(display, root, &RootIDRet, &ChildIDRet, &XCoord, &YCoord, &WinX, &WinY, &StateMask);

    /* If the position changed, return true; otherwise, return false */
    if (WinX != prevX || WinY != prevY)
    {
        /* Set the new X and Y back into Java */
        (*env)->SetIntField(env, obj, x_id, WinX);
        (*env)->SetIntField(env, obj, y_id, WinY);
        return 1;
    }

    /* The mouse position did not change */
    return 0;
}

JNIEXPORT void JNICALL Java_net_sourceforge_kjmouse_X11MousePosition_setAlwaysOnTop (JNIEnv *env, jobject obj, jobject window, jobject canvas)
{
    JAWT awt;
    JAWT_DrawingSurface* ds;
    JAWT_DrawingSurfaceInfo* dsi;
    JAWT_X11DrawingSurfaceInfo* dsi_x11;
    jboolean result;
    jint lock;
    Window xWin;
    Atom aAllowedActions, aState, aWindowType, aName;
    Atom aAllowedActionAtoms[0], aStateAtoms[3], aWindowTypeAtoms[1];
    Window parent_return, root_return;
    Window *children_return;
    XTextProperty name;
    unsigned int nchildren_return;

    /* Get the AWT */
    awt.version = JAWT_VERSION_1_3;
    result = JAWT_GetAWT(env, &awt);

    /* Get the drawing surface */
    ds = awt.GetDrawingSurface(env, canvas);
    if (ds == NULL)
        return;

    /* Lock the drawing surface */
    lock = ds->Lock(ds);
    if (lock & JAWT_LOCK_ERROR) {
        awt.FreeDrawingSurface(ds);
        return;
    }

    /* Get the drawing surface info */
    dsi = ds->GetDrawingSurfaceInfo(ds);

    /* Get the platform-specific drawing info */
    dsi_x11 = (JAWT_X11DrawingSurfaceInfo*)dsi->platformInfo;

    /* Get our window handle */
    xWin = dsi_x11->drawable;
    aName = XInternAtom(dsi_x11->display, "WM_NAME", 0);
    do {
        /* Read the WM_CLIENT property, whose existence signifies a top-level Window */
        if (XGetTextProperty(dsi_x11->display, xWin, &name, aName))
        {
            /* If the property exists, then we're done with this loop */
            if (name.encoding == XA_STRING && name.value) {
                XFree(name.value);
                break;
            } else
                XFree(name.value);
        }
        
        /* This Window wasn't it, so get our parent Window */
        if (!XQueryTree(dsi_x11->display, xWin, &root_return, &parent_return, &children_return, &nchildren_return)) {
            fprintf(stderr, "KJMouse: X11MousePosition: setAlwaysOnTop: XQueryTree failed!\n");
            ds->FreeDrawingSurfaceInfo(dsi);
            ds->Unlock(ds);
            awt.FreeDrawingSurface(ds);
            return;
        }
        xWin = parent_return;
    } while (xWin);
    
    /* In the extremely unlikely event that we could not find an ancestor Window with WM_NAME, return now */
    if (!xWin) {
        fprintf(stderr, "KJMouse: X11MousePosition: setAlwaysOnTop: No ancestor Window has a WM_NAME Property!\n");
        ds->FreeDrawingSurfaceInfo(dsi);
        ds->Unlock(ds);
        awt.FreeDrawingSurface(ds);
        return;
    }

    /* Set our window to always be on top, hidden from the pager, out of the taskbar, and sticky during viewport scrolls */
    aAllowedActions = XInternAtom(dsi_x11->display, "_NET_WM_ALLOWED_ACTIONS", 0);
    XChangeProperty(dsi_x11->display, xWin, aAllowedActions, XA_ATOM, 32, PropModeReplace, (unsigned char *) aAllowedActionAtoms, 0);
    aWindowType = XInternAtom(dsi_x11->display, "_NET_WM_WINDOW_TYPE", 0);
    aWindowTypeAtoms[0] = XInternAtom(dsi_x11->display, "_NET_WM_WINDOW_TYPE_DOCK", 0);
    XChangeProperty(dsi_x11->display, xWin, aWindowType, XA_ATOM, 32, PropModeReplace, (unsigned char *) aWindowTypeAtoms, 1);
    aState = XInternAtom(dsi_x11->display, "_NET_WM_STATE", 0);
    aStateAtoms[1] = XInternAtom(dsi_x11->display, "_NET_WM_STATE_SKIP_TASKBAR", 0);
    aStateAtoms[2] = XInternAtom(dsi_x11->display, "_NET_WM_STATE_SKIP_PAGER", 0);
    aStateAtoms[3] = XInternAtom(dsi_x11->display, "_NET_WM_STATE_STICKY", 0);
    XChangeProperty(dsi_x11->display, xWin, aState, XA_ATOM, 32, PropModeReplace, (unsigned char *) aStateAtoms, 3);
    
    /* Free the drawing surface info */
    ds->FreeDrawingSurfaceInfo(dsi);

    /* Unlock the drawing surface */
    ds->Unlock(ds);

    /* Free the drawing surface */
    awt.FreeDrawingSurface(ds);
}

#ifdef __cplusplus
}
#endif
